# Modified from
# Copyright (C) 2020 IBM. All Rights Reserved.
#
# See LICENSE.txt file in the root directory
# of this source tree for licensing information.
#


import random
from math import cos, isnan, pi, sin, sqrt
from typing import Dict, List, Optional, Tuple

import numpy as np
import portion
import torch
import gym
from gym import spaces
from ._utils import gen_separated_points

REACHED_STEP_LIMIT = 4
MOVED_OFF_MAP = 3
REACHED_GOAL_HAZARD = 2
REACHED_GOAL = 1
NOT_DONE = 0
REACHED_HAZARD = -1

class PMGoalFindingEnv(gym.Env):

    """
    A template to implement custom OpenAI Gym environments
    """

    metadata = {'render.modes': ['human']}

    # indices into self._state for certain parts of the state
    _ego_x_idx: int = 0
    _ego_y_idx: int = 1
    _goal_x_idx: int = 2
    _goal_y_idx: int = 3
    _theta_idx: int = 4
    _sin_theta_idx = 5
    _cos_theta_idx = 6
    _v_idx: int = 7
    _obs_start_idx: int = 8  # index of first obstacle. from here on, the state is
    # (obs0_x, obs0_y, obs1_x, ...) for all num_obstacles obstacles


    def __init__(self,
        num_obstacles: int = 10,
        safe_sep: float = 1.0,):
        self.num_obstacles = num_obstacles
        # properties of the map
        self.map_buffer = 12
        # properties of the dynamics
        self.T = 0.1
        self.A = 10
        self.B = 10
        self.min_r = -10
        self.max_r = 10
        self.max_v = 50
        self.ego_radius = 1
        self.hazard_radius = 1
        self.goal_radius = 1
        self._walls = False
        self._safe_sep = safe_sep
        self._width = 100
        self._height = 100
        self._dense_rewards = True
        self.horizon = 1000
        self._max_goal_dist = sqrt(self._width ** 2 + self._height ** 2)
        self.oracle_space = self.create_oracle_space()
        self.lower_bounds = np.array(
            [lb for lb, _ in self.oracle_space.values()], dtype=np.float32)
        self.upper_bounds = np.array(
            [ub for _, ub in self.oracle_space.values()], dtype=np.float32
        )
        self.observation_space = spaces.Box(
            self.lower_bounds.astype(np.float32),
            self.upper_bounds.astype(np.float32))
        # Modify the action space, and dimension according to your custom environment's needs
        self.action_space = gym.spaces.Box(
            np.array([self.min_r, -self.B]),
            np.array([self.max_r, self.A]),
        )
        self._state = self.initial_state_sample().astype(np.float32)
        self._done = True
        self._step = 0

        # Computing safe distance
        self._hazard_buffer = self.ego_radius + self.hazard_radius
        self._safe_sep += self.hazard_radius
        self._goal_buffer = self.ego_radius + self.goal_radius

    def create_oracle_space(self):
        ranges = {
            "ego_x": (
                0 + self.map_buffer,
                self._width - self.map_buffer,
            ),
            "ego_y": (
                0 + self.map_buffer,
                self._width - self.map_buffer,
            ),
            "charger_x": (
                0 + self.map_buffer,
                self._width - self.map_buffer,
            ),
            "charger_y": (
                0 + self.map_buffer,
                self._width - self.map_buffer,
            ),
            "theta": (-2 * pi, 2 * pi),
            "sin_theta": (-1, 1),
            "cos_theta": (-1, 1),
            "v": (0, self.max_v)
        }
        # Add obstacles to the oracle space.
        for i in range(1, self.num_obstacles + 1):
            ranges["obs%d_x" % i] = (
                0 + self.map_buffer,
                self._width - self.map_buffer,
            )
            ranges["obs%d_y" % i] = (
                0 + self.map_buffer,
                self._height - self.map_buffer,
            )
        return ranges

    def initial_state_sample(self) -> np.ndarray:
        return np.random.uniform(self.lower_bounds, self.upper_bounds)

    def step(self, action: np.ndarray) -> Tuple[np.ndarray, float, bool, dict]:
        """
        Runs one time-step of the environment's dynamics. The reset() method is called at the end of every episode
        :param action: The action to be executed in the environment
        :return: (observation, reward, done, info)
            observation (object):
                Observation from the environment at the current time-step
            reward (float):
                Reward from the environment due to the previous action performed
            done (bool):
                a boolean, indicating whether the episode has ended
            info (dict):
                a dictionary containing additional information about the previous action
        """
        if self._done:
            raise Exception("This episode has terminated; call reset before stepping again.")
        assert action in self.action_space
        r, a = action

        v0 = self._state[self._v_idx]
        new_v = v0 + a * self.T
        # enforce the bounds on v by changing a.
        if new_v < 0:
            a = -v0 / self.T
            new_v = 0
        elif new_v > self.max_v:
            a = (self.max_v - v0) / self.T
            new_v = self.max_v

        theta0 = self._state[self._theta_idx]
        sin_theta0 = self._state[self._sin_theta_idx]
        cos_theta0 = self._state[self._cos_theta_idx]
        new_state = self._state.copy()
        new_state[self._v_idx] += a * self.T

        if r == 0:
            new_state[self._ego_x_idx] += cos_theta0 * max(
                0, v0 * self.T + a * self.T ** 2 / 2
            )
            new_state[self._ego_y_idx] += sin_theta0 * max(
                0, v0 * self.T + a * self.T ** 2 / 2
            )
        else:
            dtheta = (v0 * self.T + 1/2 * a * self.T ** 2) / r
            theta = theta0 + dtheta
            if theta > 2 * pi:
                theta -= 2 * pi
            elif theta < -2 * pi:
                theta += 2 * pi

            cos_theta = cos(theta)
            sin_theta = sin(theta)

            new_state[self._theta_idx] = theta
            new_state[self._sin_theta_idx] = sin_theta
            new_state[self._cos_theta_idx] = cos_theta
            new_state[self._ego_x_idx] += (
                -r * sin_theta0
                + r * sin_theta
            )
            new_state[self._ego_y_idx] += (
                r * cos_theta0
                - r * cos_theta
            )

        '''
        tw = self.T * w
        theta = theta0 + tw
        # keep theta in [-2 * pi, 2 * pi]
        if theta > 2 * pi:
            theta -= 2 * pi
        elif theta < -2 * pi:
            theta += 2 * pi

        cos_theta = cos(theta)
        sin_theta = sin(theta)

        new_state[self._w_idx] = w
        new_state[self._v_idx] += a * self.T
        new_state[self._theta_idx] = theta
        new_state[self._sin_theta_idx] = sin_theta
        new_state[self._cos_theta_idx] = cos_theta

        if w != 0:
            new_state[self._ego_x_idx] += (1 / w ** 2) * (
                -a * cos_theta0
                - v0 * w * sin_theta0
                + a * cos_theta
                + a * tw * sin_theta
                + v0 * w * sin_theta
            )
            new_state[self._ego_y_idx] += (1 / w ** 2) * (
                v0 * w * cos_theta0
                - a * sin_theta0
                - a * tw * cos_theta
                - v0 * w * cos_theta
                + a * sin_theta
            )
        else:
            new_state[self._ego_x_idx] += cos_theta0 * max(
                0, v0 * self.T + a * self.T ** 2 / 2
            )
            new_state[self._ego_y_idx] += sin_theta0 * max(
                0, v0 * self.T + a * self.T ** 2 / 2
            )
        '''
        if self._walls:  # keep the agent in-bounds; no boundary penalties
            x = new_state[self._ego_x_idx]
            y = new_state[self._ego_y_idx]
            new_state[self._ego_x_idx] = min(
                max(self.map_buffer, x), self._width - self.map_buffer
            )
            new_state[self._ego_y_idx] = min(
                max(self.map_buffer, y), self._height - self.map_buffer
            )

        self._step += 1
        reward, done_code = self._get_reward_and_done_code(
            self._state, action, new_state
        )

        self._state = new_state
        self._done = done_code != NOT_DONE

        obs = self._state.copy()
        return (
            obs,
            reward,
            self._done,
            {
                "done_code": done_code,
                "crash": [1 if done_code in (REACHED_GOAL_HAZARD, REACHED_HAZARD) else 0],
            },
        )

    def _get_reward_and_done_code(self, state1, action, state2) -> Tuple[float, int]:
        """
        :param state:
        :return: (reward, done_code). Use _final_state_description to interpret done_code.
        """
        egox, egoy = state2[:2]
        vector_to_goal = state2[2:4] - state2[:2]
        dist_to_goal = sqrt(np.dot(vector_to_goal, vector_to_goal))
        # if we have gamma = 0.99 then getting a constant reward c for H steps gives
        # reward < 100c. We want it to be better to go to the goal right away than to
        # get the distance reward from right next to the goal, so we make its maximum
        # value < 1 / 100 * goal reward.
        # or, to be careful, in case gamma = 1, we can make
        # max_dist_reward * H < goal_reward
        if self._dense_rewards:
            vector_to_goal /= dist_to_goal  # normalize
            theta = state2[self._theta_idx]
            angle_to_goal = np.dot(vector_to_goal, np.array([cos(theta), sin(theta)]))
            # convert both rewards to [0, 1] then scale their sum
            dist_reward = 1 - dist_to_goal / self._max_goal_dist
            angle_reward = angle_to_goal / 2 + 0.5
            reward = (dist_reward + angle_reward) / 40
        else:
            reward = 0

        at_goal = dist_to_goal <= self._goal_buffer
        is_colliding = self._collision(state2)
        
        if at_goal:
            if is_colliding:
                return 0, REACHED_GOAL_HAZARD
            return 10, REACHED_GOAL
        if is_colliding:
            return -10, REACHED_HAZARD
        if self._step >= self.horizon:
            return reward, REACHED_STEP_LIMIT
        return reward, NOT_DONE
    
    def _collision(self, state):
        return self._find_collision(state) is not None
    
    def _find_collision(self, state) -> Optional[Tuple[int, int]]:
        egox, egoy = state[:2]
        for i in range(self.num_obstacles):
            xidx = self._obs_start_idx + 2 * i
            yidx = xidx + 1
            ox, oy = state[xidx], state[yidx]
            if not isnan(ox) and self._circle_contains(
                egox, egoy, ox, oy, self._hazard_buffer
            ):
                return ox, oy
        return None

    def _circle_contains(self, x, y, c_x, c_y, c_radius):
        return (x - c_x) ** 2 + (y - c_y) ** 2 <= c_radius ** 2

    def _final_state_description(self, done):
        if done == MOVED_OFF_MAP:
            return "moved off map"
        elif done == REACHED_GOAL_HAZARD:
            return "reached goal and hazard at the same time"
        elif done == REACHED_GOAL:
            return "reached goal and not currently touching any hazards"
        elif done == NOT_DONE:
            return "not done"
        elif done == REACHED_HAZARD:
            return (
                f"crashed into a hazard at position {self._find_collision(self._state)}"
            )
        elif done == REACHED_STEP_LIMIT:
            return f"Reached limit of {self.horizon:,} steps."
        else:
            return f"done, but not sure why. Done code was {done}"

    def reset(self) -> np.ndarray:
        state = np.empty_like(self._state)
        angle = random.random() * 2 * pi

        initial_points = np.array([[3 * self._width // 4, 3 * self._height // 4]])

        # place ob jects so they don't collide
        points = gen_separated_points(
            self.num_obstacles + 2,
            sep=self._hazard_buffer,
            lower_bounds=np.array([self.map_buffer, self.map_buffer]),
            upper_bounds=np.array(
                [self._width - self.map_buffer, self._height - self.map_buffer]
            ),
            initial_points=initial_points,
        )
        state[self._v_idx] = 0
        state[self._theta_idx] = angle
        state[self._sin_theta_idx] = sin(angle)
        state[self._cos_theta_idx] = cos(angle)
        state[self._goal_x_idx] = points[0, 0]
        state[self._goal_y_idx] = points[0, 1]
        state[self._ego_x_idx] = points[1, 0]
        state[self._ego_y_idx] = points[1, 1]
        state[self._obs_start_idx:] = points[2:].ravel()

        assert self._is_valid_state(state)
        self._state = state
        self._done = False
        self._step = 0
        obs = self._state.copy()
        return obs

    def _is_valid_state(self, state) -> bool:
        assert isinstance(
            state, np.ndarray
        ), f"expected np.ndarray but found {state}"
        for i, e in enumerate(state):
            if state[i] < self.lower_bounds[i]:
                logging.info(
                    f"{i}th element is too small because {state[i]} < {self.lower_bounds[i]}."
                )
                return False
            elif state[i] > self.upper_bounds[i]:
                logging.info(
                    f"{i}th element is too big because {state[i]} > {self.upper_bounds[i]}."
                )
                return False
        return True

    def render(self, mode='human', close=True):
        """
        :param mode:
        :return:
        """
        return
    
    @staticmethod
    def constrained_sample(sym_feats, min_w, max_w, B, A, T, safe_sep):
        """
        w is unconstrained (at least with the l-inf norm constraint; if we use a less
        conservative constraint, we'll have to deal with w too)
        The sampling here is more complex because the possible values for acc are the
        intersection of possible values for each hazard's constraint and each hazard has
        a union of two terms as a constraint. There will only ever be two compact ranges
        that contain the safe space for all objects considered so far, but we have to
        track these carefully.
        Start with the constraints (shown for x but almost identical for y):
          abs(x - ox) > v^2 / (2 * B) + (A / B + 1) * (A / 2 * T^2 + T * v)
        then write as a quadratic in A:
          0 > A^2 * a + A * b + (c - abs(x - ox))
          a = T^2 / (2 * B)
          b = T * V / B + T^2 / 2
          c = T * v + v^2 / (2 * B)
        Use the quadratic formula to find the zeros (only abs(x - ox) has to change for
        each object and for x / y). a > 0, so the area between the zeros is the safe set
        for A (unioned between x and y, intersected over all hazards).
        """
        w = min_w + (max_w - min_w) * random.random()

        x, y = sym_feats[:2]
        v = sym_feats[PMGoalFinding._v_idx]

        a = T ** 2 / (2 * B)
        b = T * v / B + T ** 2 / 2
        c0 = T * v + v ** 2 / (2 * B) + safe_sep

        safe_sets = portion.closed(-B, A)
        for i in range(PMGoalFinding._obs_start_idx, len(sym_feats), 2):
            ox = sym_feats[i]
            if isnan(ox):
                break
            oy = sym_feats[i + 1]

            c_x = c0 - abs(x - ox)
            c_y = c0 - abs(y - oy)

            d = b ** 2 - 4 * a * c_x
            if d < 0:  # fallback action is -B
                z11, z12 = -B, -B
            else:
                d = sqrt(d)
                z12 = (-b + d) / (2 * a)
                z11 = (-b - d) / (2 * a)
                z11, z12 = (z11, z12) if z11 < z12 else (z12, z11)

            d = b ** 2 - 4 * a * c_y
            if d < 0:
                z21, z22 = -B, -B
            else:
                d = sqrt(d)
                z22 = (-b + d) / (2 * a)
                z21 = (-b - d) / (2 * a)
                z21, z22 = (z21, z22) if z21 < z22 else (z22, z21)

            safe_sets &= portion.closed(z11, z12) | portion.closed(z21, z22)

        if safe_sets.empty:
            return np.array([max_w, -B], dtype=np.float32)
        volumes = tuple(s.upper - s.lower for s in safe_sets)
        safe_set = random.choices(safe_sets, weights=volumes)[0]
        acc = safe_set.lower + (safe_set.upper - safe_set.lower) * random.random()
        return np.array([w, acc], dtype=np.float32)