import numpy as np
import copy
import os
import sys
import argparse
import pickle
import matplotlib.pyplot as plt
from interval import Interval
from envs.cacc import CACCEnv
from rl import sac
import pdb


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--save-path', type=str, default='results')
    parser.add_argument('--save-file', type=str, default='results.pkl')
    parser.add_argument('--epochs', type=int, default=10)
    parser.add_argument('--steps-per-epoch', type=int, default=100)
    parser.add_argument('--seeds', nargs='+', default=[40, 41, 42])
    parser.add_argument('--rerun', action='store_true', default=False)
    return parser.parse_known_args()[0]


class DummyModel:
    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def model_monitor(*args, **kwargs):
        return True

    @staticmethod
    def control_monitor(*args, **kwargs):
        return None

    @staticmethod
    def is_safe(*args, **kwargs):
        return True

    @staticmethod
    def update(*args, **kwargs):
        pass

    @staticmethod
    def get_params():
        return []


class CACCModel:

    def __init__(self, t, a, b, eps=1e-5):
        self.T = t
        self.A = a
        self.B = b
        self.eps = eps

    def model_monitor(self, o, a, o2):
        if abs(o2[0] - o[0] - o[1] * self.T - 0.5 * a[0] * self.T ** 2) < self.eps \
                and abs(o2[1] - o[1] - a[0] * self.T) < self.eps:
            return True
        else:
            return False

    def control_monitor(self, o):
        base_interval = Interval(-self.B, self.A)

        interval_1 = copy.copy(base_interval)
        interval_1.upper_bound = min(-self.T / o[1], interval_1.upper_bound)
        param_a = self.T ** 2 / (2 * self.A)
        param_b = o[1] * self.T / self.A - 0.5 * self.T ** 2
        param_c = o[1] ** 2 / (2 * self.A) - o[0] - o[1] * self.T
        delta = param_b ** 2 - 4 * param_a * param_c
        if delta < 0:
            interval_1.lower_bound = interval_1.upper_bound
        interval_1.lower_bound = max((-param_b - np.sqrt(delta)) / (2 * param_a), interval_1.lower_bound)
        interval_1.upper_bound = min((-param_b + np.sqrt(delta)) / (2 * param_a), interval_1.upper_bound)
        interval_1_length = max(interval_1.upper_bound - interval_1.lower_bound, 0)

        interval_2 = copy.copy(base_interval)
        interval_2.lower_bound = max(-self.T / o[1], interval_1.lower_bound)
        if o[1] < 0:
            interval_2.lower_bound = max(o[1] ** 2 / (2 * o[0]), interval_2.lower_bound)
        interval_2_length = max(interval_2.upper_bound - interval_2.lower_bound, 0)

        assert interval_1_length + interval_2_length > 0
        if np.random.rand(1) <= interval_1_length / (interval_1_length + interval_2_length):
            interval = interval_1
        else:
            interval = interval_2

        a = np.random.rand(1) * (interval.upper_bound - interval.lower_bound) + interval.lower_bound
        assert self.is_safe(o, a)
        return a

    def is_safe(self, o, a):
        if o[1] + a[0] * self.T < 0:
            return o[0] + o[1] * self.T + 0.5 * a[0] * self.T ** 2 >= (o[1] + a[0] * self.T) ** 2 / (2 * self.A)
        elif o[1] < 0:
            return o[0] >= o[1] ** 2 / (2 * self.A)
        else:
            return True

    def update(self, o, a, o2, r, d, info):
        if a[0] > self.A:
            self.A = a[0]
        elif a[0] < -self.B:
            self.B = -a[0]

    def get_params(self):
        return [self.A, self.B]


def get_legend(model):
    if type(model) is DummyModel:
        return 'No model'
    if type(model) is CACCModel:
        return f'LJSC model (A={model.A}, B={model.B})'


def draw(model_list, stats_list, save_path, steps_per_epoch=200):
    plt.rcParams['font.sans-serif'] = 'Times New Roman'
    plt.rcParams.update({'figure.autolayout': True})
    plt.rc('font', size=16)
    assert len(model_list) == len(stats_list)
    # fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 5))
    # for i in range(len(model_list)):
    #     model = model_list[i]
    #     stats = stats_list[i]
    #     episode_stats_mean = np.array(stats['episode_stats_mean'])
    #     episode_stats_std = np.array(stats['episode_stats_std'])
    #     steps = np.arange(len(episode_stats_mean))
    #     reward_mean = np.cumsum(episode_stats_mean[:, 1])
    #     reward_std = episode_stats_std[:, 1]
    #     crash_mean = episode_stats_mean[:, 2].sum()
    #     label = get_legend(model)
    #     ax.plot(steps[::100], reward_mean[::100], marker='o', label=label)
    #     # ax.fill_between(steps[::100], reward_mean[::100] - reward_std[::100], reward_mean[::100] + reward_std[::100], alpha=0.3)
    #     print(label, 'average crashes:', crash_mean)
    # ax.set_xlabel('Environment Steps')
    # ax.set_ylabel('Accumulative Reward (Train)')
    # ax.legend(loc='best')
    # plt.savefig(os.path.join(save_path, 'cacc_reward_train.pdf'))
    # plt.savefig(os.path.join(save_path, 'cacc_reward_train.png'))
    # plt.close()

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 5))
    for i in range(len(model_list)):
        model = model_list[i]
        stats = stats_list[i]
        test_r_mean = np.array(stats['test_r_mean'])
        test_r_std = np.array(stats['test_r_std'])
        steps = np.arange(1, len(test_r_mean) + 1) * steps_per_epoch
        ax.plot(steps, test_r_mean, marker='o', label=get_legend(model))
        ax.fill_between(steps, test_r_mean - test_r_std, test_r_mean + test_r_std, alpha=0.3)
    ax.set_xlabel('Environment Steps')
    ax.set_ylabel('Average Episode Return (Test)')
    ax.legend(loc='best')
    plt.savefig(os.path.join(save_path, 'cacc_reward_test.pdf'))
    plt.savefig(os.path.join(save_path, 'cacc_reward_test.png'))
    plt.close()

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 5))
    ground_truth = False
    for i in range(len(model_list)):
        model = model_list[i]
        stats = stats_list[i]
        if type(model) == DummyModel or model.A == 100 and model.B == 100:
            continue
        model_params_mean = np.array(stats['model_params_mean'])
        steps = np.arange(1, len(model_params_mean) + 1)
        ax.semilogx(steps, model_params_mean[:, 0], label=get_legend(model) + '-A')
        ax.semilogx(steps, model_params_mean[:, 1], label=get_legend(model) + '-B')
        if not ground_truth:
            ax.plot(steps, 100 * np.ones(len(steps)), 'r--', label='Ground Truth')
            ground_truth = True
    ax.set_xlabel('Environment Steps')
    ax.set_ylabel('Model Parameters')
    ax.legend(loc='best')
    plt.savefig(os.path.join(save_path, 'cacc_model_params.pdf'))
    plt.savefig(os.path.join(save_path, 'cacc_model_params.png'))
    plt.close()


def merge_stats(stats_list):
    merged_stats = dict()
    keys = stats_list[0].keys()
    for key in keys:
        value = [stats[key] for stats in stats_list]
        value = np.stack(value, axis=0)
        merged_stats[key + '_mean'] = np.mean(value, axis=0)
        merged_stats[key + '_std'] = np.std(value, axis=0)
    return merged_stats


def process_data(model_list, stats_list):
    legends = []
    processed_model_list = []
    processed_stats_list = []
    for model in model_list:
        legends.append(get_legend(model))
    legends = list(set(legends))
    for legend in legends:
        matching_num = 0
        matching_stats = []
        for i in range(len(model_list)):
            model = model_list[i]
            stats = stats_list[i]
            if get_legend(model) != legend:
                continue
            matching_num += 1
            if matching_num == 1:
                processed_model_list.append(model)
            matching_stats.append(stats)
        processed_stats_list.append(merge_stats(matching_stats))
    return processed_model_list, processed_stats_list


def main(args):
    env = CACCEnv
    save_file_path = os.path.join(args.save_path, args.save_file)
    if not os.path.isfile(save_file_path) or args.rerun:
        model_list = [DummyModel(), CACCModel(0.1, 10, 10), CACCModel(0.1, 100, 100)]
        stats_list = []
        for seed in args.seeds:
            for model in model_list:
                stats, _ = sac(env, device='cpu', model=copy.copy(model), steps_per_epoch=args.steps_per_epoch,
                               epochs=args.epochs, seed=seed)
                stats_list.append(stats)
        model_list *= len(args.seeds)
        data = (model_list, stats_list)
        with open(save_file_path, 'wb') as f:
            pickle.dump(data, f)
    else:
        with open(save_file_path, 'rb') as f:
            data = pickle.load(f)
    draw(*process_data(*data), args.save_path, args.steps_per_epoch)


if __name__ == '__main__':
    main(parse_args())
