import numpy as np
import copy
import os
import sys
import matplotlib.pyplot as plt
from interval import Interval
from envs.pm_goal_finding import PMGoalFindingEnv
from envs.pointmass import PointmassEnv
from rl import sac
import pdb
import random
import pickle
import argparse

MAX_ITERS = 100

def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--save-path', type=str, default='results')
    parser.add_argument('--save-file', type=str, default='results.pkl')
    # env == "GoalFinding" or "Pointmess" to set up different environments
    parser.add_argument('--env', type=str, default="GoalFinding")
    parser.add_argument('--epochs', type=int, default=5)
    parser.add_argument('--steps-per-epoch', type=int, default=2000)
    parser.add_argument('--seeds', nargs='+', default=[0,1,2])
    # default to run stored model
    parser.add_argument('--rerun', action='store_true', default=False)
    return parser.parse_known_args()[0]

class DummyModel:
    def __init__(self, *args, **kwargs):
        pass

    @staticmethod
    def model_monitor(*args, **kwargs):
        return True

    @staticmethod
    def control_monitor(*args, **kwargs):
        return None

    @staticmethod
    def is_safe(*args, **kwargs):
        return True

    @staticmethod
    def update(*args, **kwargs):
        pass

    @staticmethod
    def get_params():
        return []


class PMGoalFindingModel:

    _ego_x_idx: int = 0
    _ego_y_idx: int = 1
    _goal_x_idx: int = 2
    _goal_y_idx: int = 3
    _theta_idx: int = 4
    _sin_theta_idx = 5
    _cos_theta_idx = 6
    _v_idx: int = 7
    _obs_start_idx: int = 8  # index of first obstacle. from here on, the state is
    # (obs0_x, obs0_y, obs1_x, ...) for all num_obstacles obstacles
    # sync this with number of obstacles as defined in pointmass
    num_obstacles = 10

    def __init__(self, min_r, max_r, A, B):
        self.min_r = min_r
        self.max_r = max_r
        self.A = A
        self.B = B
        self.T = 0.1
        self.eps = 1e-3
        self._hazard_buffer = 2

    def model_monitor(self, obs, action, obs2):
        r,a = action
        ego_x, ego_y = obs[self._ego_x_idx], obs[self._ego_y_idx]
        ego_x_post, ego_y_post = obs2[self._ego_x_idx], obs2[self._ego_y_idx]
        if r != 0:
            if abs(ego_x - r * obs[self._sin_theta_idx] + r * obs2[self._sin_theta_idx] - ego_x_post) > self.eps:
                print(abs(ego_x - r * obs[self._sin_theta_idx] + r * obs2[self._sin_theta_idx] - ego_x_post))
                return False
            if abs(ego_y + r * obs[self._cos_theta_idx] - r * obs2[self._cos_theta_idx] - ego_y_post) > self.eps:
                print(abs(ego_y + r * obs[self._cos_theta_idx] - r * obs2[self._cos_theta_idx] - ego_y_post))
                return False
        return True

    # Find a suitable radius r and acceleration a
    def control_monitor(self, obs):
        (low, high) = (self.min_r, self.max_r)
        acc = self.A
        # acc = random.random() * (high_a - low_a) + low_a
        r = random.random() * (high - low) + low
        act = [r, acc]
        num_iters = 0
        num_obs = len(obs[self._obs_start_idx:]) // 2
        while not self.is_safe(obs, act):
            # acc = random.random() * (high_a - low_a) + low_a
            r = random.random() * (high - low) + low
            act = [r, acc]
            print(obs)
            print(act)
            num_iters += 1
            if num_iters >= MAX_ITERS:
                print("crash")
                return act
        return act

    # Check if the obsveration is safe by performing action
    def is_safe(self, obs, action):
        ego_x = obs[self._ego_x_idx]
        ego_y = obs[self._ego_y_idx]
        v = obs[self._v_idx]
        theta = obs[self._theta_idx]
        sin_theta = obs[self._sin_theta_idx]
        cos_theta = obs[self._cos_theta_idx]
        num_obs = self.num_obstacles
        r,a = action
        obs_x = []
        obs_y = []
        for i in range(num_obs):
            obs_x.append(obs[self._obs_start_idx+i*2])
            obs_y.append(obs[self._obs_start_idx+i*2+1])
        # check collision with each distinct obstacle
        for i in range(num_obs):
            if r > 0:
                if (ego_x - r*sin_theta - obs_x[i]) ** 2 + \
                (ego_y + r*cos_theta - obs_y[i]) ** 2 < (r + self._hazard_buffer) ** 2 and \
                (ego_x - r*sin_theta - obs_x[i]) ** 2 + \
                (ego_y + r*cos_theta - obs_y[i]) ** 2 > (r - self._hazard_buffer) ** 2:
                    return False
            elif r < 0:
                if (ego_x - r*sin_theta - obs_x[i]) ** 2 + \
                (ego_y + r*cos_theta - obs_y[i]) ** 2 < (r - self._hazard_buffer) ** 2 and \
                (ego_x - r*sin_theta - obs_x[i]) ** 2 + \
                (ego_y + r*cos_theta - obs_y[i]) ** 2 > (r + self._hazard_buffer) ** 2:
                    return False
            else:
                if (obs_x[i] - (ego_x + cos_theta * max(0, v * self.T + a * self.T ** 2 / 2)) < self._hazard_buffer or obs_y[i] - (ego_y + sin_theta * max(0, v * self.T + a * self.T ** 2 / 2)) < self._hazard_buffer):
                    return False
        return True

    def update(self, o, a, o2, r, d, info):
        if a[0] > self.max_r:
            self.max_r = a[0]
        elif a[0] < self.min_r:
            self.min_r = a[0]
        
        if a[1] > self.A:
            self.A = a[1]
        elif a[1] < -self.B:
            self.B = -a[1]
        print(self.get_params())

    def get_params(self):
        return [(self.min_r, self.max_r)]

def get_legend(model,index=0, min=True):
    if type(model) is DummyModel:
        return 'No model'
    if type(model) is PMGoalFindingModel:
        return f'LJSC model (min_r={model.min_r}, max_r={model.max_r})'

def draw(model_list, stats_list, save_path, steps_per_epoch=200):
    plt.rcParams['font.sans-serif'] = 'Times New Roman'
    plt.rcParams.update({'figure.autolayout': True})
    plt.rc('font', size=16)
    assert len(model_list) == len(stats_list)
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 5))
    for i in range(len(model_list)):
        model = model_list[i]
        stats = stats_list[i]
        
        episode_stats_mean = np.array(stats['episode_stats_mean'])
        print([i for i in episode_stats_mean[:, 2] if i != 0])
        episode_stats_std = np.array(stats['episode_stats_std'])
        steps = np.arange(len(episode_stats_mean))
        reward_mean = np.cumsum(episode_stats_mean[:, 1])
        reward_std = episode_stats_std[:, 1]
        crash_mean = episode_stats_mean[:, 2].sum()
        label = get_legend(model)
        ax.plot(steps[::100], reward_mean[::100], marker='o', label=label)
        # ax.fill_between(steps[::100], reward_mean[::100] - reward_std[::100], reward_mean[::100] + reward_std[::100], alpha=0.3)
        print(label, 'average crashes:', crash_mean)
    ax.set_xlabel('Environment Steps')
    ax.set_ylabel('Accumulative Reward (Train)')
    ax.legend(loc='best')
    plt.savefig(os.path.join(save_path, 'pm_reward_train.pdf'))
    plt.savefig(os.path.join(save_path, 'pm_reward_train.png'))
    plt.close()

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 5))
    for i in range(len(model_list)):
        model = model_list[i]
        stats = stats_list[i]
        test_r_mean = np.array(stats['test_r_mean'])
        test_r_std = np.array(stats['test_r_std'])
        steps = np.arange(1, len(test_r_mean) + 1) * steps_per_epoch
        ax.plot(steps, test_r_mean, marker='o', label=get_legend(model))
        ax.fill_between(steps, test_r_mean - test_r_std, test_r_mean + test_r_std, alpha=0.3)
    ax.set_xlabel('Environment Steps')
    ax.set_ylabel('Average Episode Return (Test)')
    ax.legend(loc='best')
    plt.savefig(os.path.join(save_path, 'pm_reward_test.pdf'))
    plt.savefig(os.path.join(save_path, 'pm_reward_test.png'))
    plt.close()

    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(8, 5))
    ground_truth = False
    for i in range(len(model_list)):
        model = model_list[i]
        stats = stats_list[i]
        if type(model) == DummyModel: # or model.A == 100 and model.B == 100:
            continue
        model_params_mean = np.array(stats['model_params_mean'])
        steps = np.arange(1, len(model_params_mean) + 1)
        ax.semilogx(steps, model_params_mean[:, 0, 0], label=get_legend(model) + '-max_r')
        ax.semilogx(steps, model_params_mean[:, 0, 1], label=get_legend(model) + '-min_r')
        if not ground_truth:
            ax.plot(steps, 10 * np.ones(len(steps)), 'r--', label='Ground Truth')
            ground_truth = True
    ax.set_xlabel('Environment Steps')
    ax.set_ylabel('Model Parameters')
    ax.legend(loc='best')
    plt.savefig(os.path.join(save_path, 'pm_model_params.pdf'))
    plt.savefig(os.path.join(save_path, 'pm_model_params.png'))
    plt.close()

def merge_stats(stats_list):
    merged_stats = dict()
    keys = stats_list[0].keys()
    for key in keys:
        value = [stats[key] for stats in stats_list]
        value = np.stack(value, axis=0)
        merged_stats[key + '_mean'] = np.mean(value, axis=0)
        merged_stats[key + '_std'] = np.std(value, axis=0)
    return merged_stats

def process_data(model_list, stats_list):
    legends = []
    processed_model_list = []
    processed_stats_list = []
    for model in model_list:
        legends.append(get_legend(model))
    legends = list(set(legends))
    for legend in legends:
        matching_num = 0
        matching_stats = []
        for i in range(len(model_list)):
            model = model_list[i]
            stats = stats_list[i]
            if get_legend(model) != legend:
                continue
            matching_num += 1
            if matching_num == 1:
                processed_model_list.append(model)
            matching_stats.append(stats)
        processed_stats_list.append(merge_stats(matching_stats))
    return processed_model_list, processed_stats_list

def main(args):
    save_file_path = os.path.join(args.save_path, args.save_file)
    if args.env == "GoalFinding":
        env = PMGoalFindingEnv
    else:
        env = PointmassEnv
    if not os.path.isfile(save_file_path) or args.rerun:
        model_list = [DummyModel(), PMGoalFindingModel(-1,1, 10, 10)]
        stats_list = []
        for seed in args.seeds:
            for model in model_list:
                stats, _ = sac(env, device='cpu', model=copy.copy(model), steps_per_epoch=args.steps_per_epoch, epochs=args.epochs, seed=seed, max_ep_len=1000)
                stats_list.append(stats)
        model_list *= len(args.seeds)
        data = (model_list, stats_list)
        with open(save_file_path, 'wb') as f:
            pickle.dump(data, f)
    else:
        with open(save_file_path, 'rb') as f:
            data = pickle.load(f)
            print(len(data[1]))
    draw(*process_data(*data), args.save_path, args.steps_per_epoch)

if __name__ == '__main__':
    main(parse_args())
